import numpy as np
import matplotlib.pyplot as plt
import pystan
import pickle
import seaborn as sb
import csv
import scipy.optimize as sciopt
from my_module2004 import *

fin_nm = input('Data file name (*.csv) = ')
fout_nm = input('Output file name (*.txt) = ')
fout = open(fout_nm, 'w')
fout.write('Data file...{}\n'.format(fin_nm))
#
#       Read data from CSV file
#
with open(fin_nm, 'r') as fin:
    data = [v for v in csv.reader(fin)]
N = len(data) - 1
ID = []
X = np.empty(N, dtype = float)
Res = np.empty(N, dtype = int)
for t in range(N):
    ID.append(data[t + 1][0])
    X[t] = float(data[t + 1][1])
    Res[t] = int(data[t + 1][2])

print('Number of data = ', N)
fout.write('\nNumber of data = {}\n'.format(N))
for i in range(N):
    print('{0}  {1}  {2}'.format(ID[i], X[i], Res[i]))

for i, s, r in zip(ID, X, Res):
    fout.write('{0:>5s}{1:>7.2f}{2:>5d}\n'.format(i, s, r))
#
#       Boundaries and initial values
#

C1C2 = np.percentile(X, [20, 80])

param_a = 1.0           #   Parameter a of a gamma distribution for the prior distribution 
param_b = 0.001         #   Parameter b of a gamma distribution for the prior distribution 

def f_init():
    """
                Initial values for pse(mu) and sigma
    """
    return dict(pse = (C1C2[0] + C1C2[1]) / 2, sgm = C1C2[1] - C1C2[0])

stan_data = {'N': N, 'X':X, 'Res': Res, 'a': param_a, 'b': param_b}
#
#       Sampling by Stan
#
sm = pystan.StanModel(file = 'PM2Cat2004.stan')

fit = sm.sampling(data = stan_data, pars = ['sgm', 'pse'], init = f_init, n_jobs = 1)

print(fit)
fout.write('\nSummary...\n{}\n'.format(fit))

SmplSgm = fit['sgm']        #   Samples from the posterior distribution for sigma
SmplPSE = fit['pse']        #   Samples from the posterior distribution for pse(mu)
#
#       Percentile points for posterior distributions
#
SgmL025 = np.percentile(SmplSgm, 2.5)
SgmMed = np.percentile(SmplSgm, 50)
SgmU975 = np.percentile(SmplSgm, 97.5)
PSEL025 = np.percentile(SmplPSE, 2.5)
PSEMed = np.percentile(SmplPSE, 50)
PSEU975 = np.percentile(SmplPSE, 97.5)
#
#       Calculate MAP estimates for mu(pse) and sigma
#
#           calcMAP from my_module.py
#
muMAP, sgmMAP = calcMAP(X, Res, PSEMed, SgmMed, param_a, param_b)    
print('mu = ', muMAP, '    sgm = ', sgmMAP)
fout.write('\nmu(MAP) = {0:<.3f}    sigma(MAP) = {1:<.3f}\n'.format(muMAP, sgmMAP))
#
#       KDE plot for the samples from posterior distribution for PSE(mu)
#
sb.kdeplot(SmplPSE)
plt.plot([PSEL025, PSEU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(muMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\mu$', fontsize = 14)
plt.title('Posterior Distribution for $\mu$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(muMAP, PSEL025, PSEU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       KDE plot for the samples from posterior distribution for sigma
#
sb.kdeplot(SmplSgm)
plt.plot([SgmL025, SgmU975], [0.0, 0.0], color = 'b', linewidth = 5, label = '95% CI')
plt.plot(sgmMAP, 0.0, 'go', markersize = 10, label = 'MAP')
plt.yticks([])
plt.xlabel('$\sigma$', fontsize = 14)
plt.title('Posterior Distribution for $\sigma$\n' +
          ('MAP = {0:<.2f}, 95% CI = [{1:<.2f}, {2:<.2f}]'.format(sgmMAP, SgmL025, SgmU975)),
          fontsize = 16)
plt.legend()
plt.tight_layout(h_pad = 5)
plt.show()
#
#       Prob(z < s75) = 0.75 for the standard normal distribution
#
s75 = bisect(f_phi, 0.75, 0.5, 1.0)             #   f_phi from my_module.py
print('s75 = ', s75, '     p = ', f_phi(s75))
JND = sgmMAP * s75
PSE = muMAP
fout.write('\nPSE = {0:<.3f},    95% CI = [{1:<.2f}, {2:<.2f}]'.format(
            PSE, PSEL025, PSEU975))
fout.write('\nJND = {0:<.2f},    95% CI = [{1:<.2f}, {2:<.2f}]'.format(
            JND, s75 * SgmL025, s75 * SgmU975))
#
#       Data and the psychometric function
#
pf = PF(muMAP, sgmMAP)          #   Class PF from my_module.py
minX = X[np.argmin(X)]
maxX = X[np.argmax(X)]
xvalues = []
yvalues = []
for i in range(1001):
    t = minX + i * (maxX - minX) / 1000.0
    xvalues.append(t)
    yvalues.append(pf.v(t))     #   Values of the psychometric function
plt.title('Psychometric Function\n' +
          ('PSE = {0:<.2f},  JND = {1:<.2f}'.format(PSE, JND)),
          fontsize = 16)
plt.plot(xvalues, yvalues, linewidth = 3, label = 'PF')
plt.yticks([0.0, 0.5, 0.75, 1.0])
#
#       Resposes are scaled to values from 0 to 1, and added small fluctuations
#
Res_rn = (Res - 1) + np.random.uniform(-0.05, 0.05, len(Res))   
plt.scatter(X, Res_rn, alpha = 0.5, color = 'g', label = 'Data')
plt.plot([minX, maxX], [0.0, 0.0], color = 'k')
plt.plot([minX, maxX], [1.0, 1.0], color = 'k')
plt.plot([minX, PSE, PSE], [0.5, 0.5, 0.0], color = 'k')
plt.plot([minX, PSE + JND, PSE + JND], [0.75, 0.75, 0.0], color = 'k')
plt.legend(loc = 7)
plt.xlabel('Stimulus value', fontsize = 12)
plt.ylabel('Probability', fontsize = 12)
plt.show()

#
#       Observed responses smoothed by Kernel density estimation 
#
w_1 = np.empty(len(Res))        #   Weights for Response >= 2
for t in range(len(Res)):
    w_1[t] = 0 if Res[t] < 2 else 1

X_range = np.arange(minX, maxX, (maxX - minX) / 100)
w_k = JND / 2

while True:
    pm_1 = []           #   KDE values for Response >= 2
    for v in X_range:
        pm_1.append(k_ave(v, X, w_1, w_k))
        
    plt.title('Observed and Estimated PFs\n' +
              ('PSE = {0:<.2f},  JND = {1:<.2f}'.format(PSE, JND)),
              fontsize = 16)
    plt.plot(xvalues, yvalues, color = 'b', linewidth = 3, label = 'PF')
    plt.yticks([0.0, 0.5, 1.0])
    plt.plot(X_range, pm_1, 'g--', linewidth = 2, label = 'Res(>1)')
    plt.xlabel('Stimulus value', fontsize = 12)
    plt.legend(loc = 'upper left')
    plt.show()
    print('\nThe current value of w_k = {}'.format(w_k))
    w_k = float(input('The next value of w_k = '))
    if w_k <= 0.0:
        break

fout.close()

print('\nOutput file {} was saved.'.format(fout_nm))
